#
#   Copyright 2023 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import base64
import datetime
import json
import logging
import warnings
from functools import wraps
from typing import Any, Dict, Optional, Union

from hopsworks_common.core.constants import HAS_PYARROW, pyarrow_not_installed_message


if not HAS_PYARROW:
    raise ModuleNotFoundError(pyarrow_not_installed_message)

import pyarrow
import pyarrow._flight
import pyarrow.flight
from hopsworks_common import client
from hopsworks_common.client.exceptions import FeatureStoreException
from hopsworks_common.core.constants import HAS_POLARS, polars_not_installed_message
from hsfs import feature_group
from hsfs.constructor import query
from hsfs.core.variable_api import VariableApi
from hsfs.storage_connector import StorageConnector
from pyarrow.flight import FlightServerError
from retrying import retry


if HAS_POLARS:
    import polars as pl


_logger = logging.getLogger(__name__)


_arrow_flight_instance = None


def get_instance() -> ArrowFlightClient:
    global _arrow_flight_instance
    if not _arrow_flight_instance:
        _arrow_flight_instance = ArrowFlightClient()
    return _arrow_flight_instance


def close() -> None:
    global _arrow_flight_instance
    _arrow_flight_instance = None


def _disable_feature_query_service_client():
    global _arrow_flight_instance
    _logger.debug("Disabling Hopsworks Query Service Client.")
    if _arrow_flight_instance is None:
        _arrow_flight_instance.ArrowFlightClient(disabled_for_session=True)
    else:
        _arrow_flight_instance._disable_for_session(on_purpose=True)


def _should_retry(exception):
    return isinstance(exception, pyarrow._flight.FlightUnavailableError)


def _is_feature_query_service_queue_full_error(exception):
    return isinstance(
        exception, pyarrow._flight.FlightServerError
    ) and "no free slot available for" in str(exception)


def _is_no_commits_found_error(exception):
    return isinstance(
        exception, pyarrow._flight.FlightServerError
    ) and "No commits found" in str(exception)


def _should_retry_healthcheck(exception):
    return (
        isinstance(exception, pyarrow._flight.FlightUnavailableError)
        or isinstance(exception, pyarrow._flight.FlightTimedOutError)
    )


def _should_retry_certificate_registration(exception):
    return (
        _should_retry_healthcheck(exception)
        or _is_feature_query_service_queue_full_error(exception)
    )


# Avoid unnecessary client init
def is_data_format_supported(data_format: str, read_options: Optional[Dict[str, Any]]):
    if data_format not in ArrowFlightClient.SUPPORTED_FORMATS:
        return False
    elif read_options and read_options.get("use_spark", False):
        return False
    else:
        return get_instance()._should_be_used()


def _is_query_supported_rec(query: query.Query):
    hudi_no_time_travel = (
        isinstance(query._left_feature_group, feature_group.FeatureGroup)
        and query._left_feature_group.time_travel_format == "HUDI"
        and (
            query._left_feature_group_start_time is None
            or query._left_feature_group_start_time == 0
        )
        and query._left_feature_group_end_time is None
    )
    supported_connector = (
        isinstance(query._left_feature_group, feature_group.ExternalFeatureGroup)
        and query._left_feature_group.storage_connector.type
        in ArrowFlightClient.SUPPORTED_EXTERNAL_CONNECTORS
    )
    delta_s3 = (
        isinstance(query._left_feature_group, feature_group.FeatureGroup)
        and query._left_feature_group.time_travel_format == "DELTA"
        and query._left_feature_group.storage_connector
        and query._left_feature_group.storage_connector.type == StorageConnector.S3
    )
    supported = hudi_no_time_travel or supported_connector or delta_s3
    for j in query._joins:
        supported &= _is_query_supported_rec(j._query)
    return supported


def is_query_supported(query: query.Query, read_options: Optional[Dict[str, Any]]):
    if read_options and read_options.get("use_spark", False):
        return False
    elif not _is_query_supported_rec(query):
        return False
    else:
        return get_instance()._should_be_used()


class ArrowFlightClient:
    SUPPORTED_FORMATS = ["parquet"]
    SUPPORTED_EXTERNAL_CONNECTORS = [
        StorageConnector.SNOWFLAKE,
        StorageConnector.BIGQUERY,
    ]
    READ_ERROR = "Could not read data using Hopsworks Query Service."
    WRITE_ERROR = 'Could not write data using Hopsworks Query Service. If the issue persists, use write_options={"use_spark": True} instead.'
    DEFAULTING_TO_DIFFERENT_SERVICE_WARNING = (
        "Defaulting to Spark execution for this call."
    )
    CLIENT_WILL_STAY_ACTIVE_WARNING = 'The client will remain active for future calls. If the issue persists write_options={"use_spark": True}.'
    DEFAULT_TIMEOUT_SECONDS = 900
    DEFAULT_HEALTHCHECK_TIMEOUT_SECONDS = 5
    DEFAULT_GRPC_MIN_RECONNECT_BACKOFF_MS = 2000

    def __init__(self, disabled_for_session: bool = False):
        _logger.debug("Initializing Hopsworks Query Service Client.")
        self._timeout: float = ArrowFlightClient.DEFAULT_TIMEOUT_SECONDS
        self._health_check_timeout: float = (
            ArrowFlightClient.DEFAULT_HEALTHCHECK_TIMEOUT_SECONDS
        )

        self._enabled_on_cluster: bool = False
        self._host_url: Optional[str] = None
        self._connection: Optional[pyarrow.flight.FlightClient] = None
        if disabled_for_session:
            self._disable_for_session(on_purpose=True)
            return
        else:
            self._disabled_for_session: bool = False

        self._client = client.get_instance()
        self._variable_api: VariableApi = VariableApi()
        self._certificates_json: Optional[str] = None

        try:
            self._check_cluster_service_enabled()
            self._host_url = self._retrieve_host_url()

            if self._enabled_on_cluster:
                _logger.debug("Hopsworks Query Service is enabled on the cluster.")
                self._initialize_flight_client()
            else:
                _logger.debug(
                    "Hopsworks Query Service Client is not enabled on the cluster or a cluster variable is misconfigured."
                )
                self._disable_for_session()
                return
        except Exception as e:
            _logger.debug("Failed to connect to Hopsworks Query Service")
            _logger.exception(e)
            self._disable_for_session(str(e))
            return

        try:
            self._health_check()
            if "get-version" in [action.type for action in self._connection.list_actions()]:
                self._server_version = self._get_server_version()
            else:
                self._server_version = None
            if self._server_version is None:
                self._register_certificates()
        except Exception as e:
            _logger.debug("Failed to connect to Hopsworks Query Service.")
            _logger.exception(e)
            warnings.warn(
                f"Failed to connect to Hopsworks Query Service, got {str(e)}."
                + ArrowFlightClient.DEFAULTING_TO_DIFFERENT_SERVICE_WARNING
                + ArrowFlightClient.CLIENT_WILL_STAY_ACTIVE_WARNING,
                stacklevel=1,
            )
            return

    def _check_cluster_service_enabled(self) -> None:
        try:
            _logger.debug(
                "Connecting to Hopsworks Cluster to check if Hopsworks Query Service is enabled."
            )
            self._enabled_on_cluster = self._variable_api.get_flyingduck_enabled()
        except Exception as e:
            # if feature flag cannot be retrieved, assume it is disabled
            _logger.debug(
                "Unable to fetch Hopsworks Query Service flag, disabling its client."
            )
            _logger.exception(e)
            self._enabled_on_cluster = False

    def _retrieve_host_url(self) -> Optional[str]:
        _logger.debug("Retrieving host URL.")
        if client._is_external():
            external_domain = self._variable_api.get_loadbalancer_external_domain(
                "feature_query"
            )
            host_url = f"grpc+tls://{external_domain}:5005"
        else:
            service_discovery_domain = self._variable_api.get_service_discovery_domain()
            if service_discovery_domain == "":
                raise FeatureStoreException(
                    "Client could not get Hopsworks Query Service hostname from service_discovery_domain. "
                    "The variable is either not set or empty in Hopsworks cluster configuration."
                )
            host_url = f"grpc+tls://flyingduck.service.{service_discovery_domain}:5005"
        _logger.debug(f"Connecting to Hopsworks Query Service on host {host_url}")
        return host_url

    def _disable_for_session(
        self, message: Optional[str] = None, on_purpose: bool = False
    ) -> None:
        self._disabled_for_session = True
        if on_purpose:
            warnings.warn(
                "Hopsworks Query Service will be disabled for this session.",
                stacklevel=1,
            )
        if self._enabled_on_cluster:
            warnings.warn(
                "Hospworks Query Service is disabled on cluster. Contact your administrator to enable it.",
                stacklevel=1,
            )
        else:
            warnings.warn(
                f"Client initialisation failed: {message}. Hopsworks Query Service will be disabled for this session."
                "If you believe this is a transient error, you can call `(hopsworks.)hsfs.reset_offline_query_service_client()`"
                " to re-enable it.",
                stacklevel=1,
            )

    def _initialize_flight_client(self):
        (tls_root_certs, cert_chain, private_key) = self._extract_certs()
        self._connection = pyarrow.flight.FlightClient(
            location=self.host_url,
            tls_root_certs=tls_root_certs,
            cert_chain=cert_chain,
            private_key=private_key,
            override_hostname="flyingduck.service.consul",
            generic_options=[
                (
                    # https://arrow.apache.org/docs/cpp/flight.html#excessive-traffic
                    "GRPC_ARG_MIN_RECONNECT_BACKOFF_MS",
                    ArrowFlightClient.DEFAULT_GRPC_MIN_RECONNECT_BACKOFF_MS,
                )
            ],
        )

    @retry(
        wait_exponential_multiplier=1000,
        stop_max_attempt_number=5,
        retry_on_exception=_should_retry_healthcheck,
    )
    def _health_check(self):
        _logger.debug("Performing healthcheck of Hopsworks Query Service.")
        action = pyarrow.flight.Action("healthcheck", b"")
        options = pyarrow.flight.FlightCallOptions(timeout=self.health_check_timeout)
        list(self._connection.do_action(action, options=options))
        _logger.debug("Healthcheck succeeded.")

    @retry(
        wait_exponential_multiplier=1000,
        stop_max_attempt_number=3,
        retry_on_exception=_should_retry,
    )
    def _get_server_version(self):
        _logger.debug("Acquiring the server version of Hopsworks Query Service.")
        action = pyarrow.flight.Action("get-version", b"")
        options = pyarrow.flight.FlightCallOptions(timeout=self.health_check_timeout)
        for res in self._connection.do_action(action, options=options):
            version = res.body.to_pybytes()
            _logger.debug(f"The HQS server is of version {version}.")
            return version

    def _should_be_used(self):
        if not self._enabled_on_cluster:
            _logger.debug(
                "Hopsworks Query Service not used as it is disabled on the cluster."
            )
            return False

        if self._disabled_for_session:
            _logger.debug(
                "Hopsworks Query Service client failed to initialise and is disabled for the session."
            )
            return False

        _logger.debug("Using Hopsworks Query Service.")
        return True

    def _extract_certs(self):
        _logger.debug("Extracting client certificates.")
        with open(self._client._get_ca_chain_path(), "rb") as f:
            tls_root_certs = f.read()
        with open(self._client._get_client_cert_path(), "r") as f:
            cert_chain = f.read()
        with open(self._client._get_client_key_path(), "r") as f:
            private_key = f.read()
        return tls_root_certs, cert_chain, private_key

    def _encode_certs(self, path):
        _logger.debug(f"Encoding certificates from path: {path}")
        with open(path, "rb") as f:
            content = f.read()
            return base64.b64encode(content).decode("utf-8")

    def _certificates(self):
        kstore = self._encode_certs(self._client._get_jks_key_store_path())
        tstore = self._encode_certs(self._client._get_jks_trust_store_path())
        cert_key = self._client._cert_key
        return {"kstore": kstore, "tstore": tstore, "cert_key": cert_key}

    def _make_certificates_json(self):
        if self._certificates_json is None:
            self._certificates_json = json.dumps(self._certificates()).encode("utf-8")
        return self._certificates_json

    def _certificates_headers(self):
        if self._server_version is None:
            return []
        return [(b"x-certificates-json", self._make_certificates_json())]

    @retry(
        wait_exponential_multiplier=1000,
        stop_max_attempt_number=3,
        retry_on_exception=_should_retry_certificate_registration,
    )
    def _register_certificates(self):
        certificates_json = self._make_certificates_json()
        certificates_json_buf = pyarrow.py_buffer(certificates_json)
        action = pyarrow.flight.Action(
            "register-client-certificates", certificates_json_buf
        )
        # Registering certificates queue time occasionally spike.
        options = pyarrow.flight.FlightCallOptions(timeout=self.health_check_timeout)
        _logger.debug("Registering client certificates with Hopsworks Query Service.")
        self._connection.do_action(action, options=options)
        _logger.debug("Client certificates registered.")

    def _handle_afs_exception(user_message="None"):
        def decorator(func):
            @wraps(func)
            def afs_error_handler_wrapper(instance, *args, **kw):
                try:
                    return func(instance, *args, **kw)
                except Exception as e:
                    message = str(e)
                    _logger.debug("Caught exception in %s: %s", func.__name__, message)
                    _logger.exception(e)
                    if instance._server_version is None and (
                        isinstance(e, FlightServerError)
                        and "Please register client certificates first." in message
                    ):
                        instance._register_certificates()
                        return func(instance, *args, **kw)
                    elif _is_feature_query_service_queue_full_error(e):
                        raise FeatureStoreException(
                            "Hopsworks Query Service is busy right now. Please try again later."
                        ) from e
                    elif _is_no_commits_found_error(e):
                        raise FeatureStoreException(str(e).split("Details:")[0]) from e
                    else:
                        raise FeatureStoreException(user_message) from e

            return afs_error_handler_wrapper

        return decorator

    @retry(
        wait_exponential_multiplier=1000,
        stop_max_attempt_number=3,
        retry_on_exception=_should_retry,
    )
    def get_flight_info(self, descriptor):
        # The timeout needs not be as long as timeout for do_get or do_action
        _logger.debug("Getting flight info for descriptor: %s", str(descriptor))
        options = pyarrow.flight.FlightCallOptions(timeout=self.health_check_timeout)
        return self._connection.get_flight_info(
            descriptor,
            options=options,
        )

    @retry(
        wait_exponential_multiplier=1000,
        stop_max_attempt_number=3,
        retry_on_exception=_should_retry,
    )
    def _get_dataset(self, descriptor, timeout=None, dataframe_type="pandas"):
        if timeout is None:
            timeout = self.timeout
        info = self.get_flight_info(descriptor)
        _logger.debug("Retrieved flight info: %s. Fetching dataset.", str(info))
        options = pyarrow.flight.FlightCallOptions(
            timeout=timeout, headers=self._certificates_headers()
        )
        reader = self._connection.do_get(info.endpoints[0].ticket, options)
        _logger.debug("Dataset fetched. Converting to dataframe %s.", dataframe_type)
        if dataframe_type.lower() == "polars":
            if not HAS_POLARS:
                raise ModuleNotFoundError(polars_not_installed_message)
            return pl.from_arrow(reader.read_all())
        else:
            return reader.read_pandas()

    # retry is handled in get_dataset
    @_handle_afs_exception(user_message=READ_ERROR)
    def read_query(self, query_object, arrow_flight_config, dataframe_type):
        query_encoded = json.dumps(query_object).encode("ascii")
        descriptor = pyarrow.flight.FlightDescriptor.for_command(query_encoded)
        return self._get_dataset(
            descriptor,
            (
                arrow_flight_config.get("timeout", self.timeout)
                if arrow_flight_config
                else self.timeout
            ),
            dataframe_type,
        )

    # retry is handled in get_dataset
    @_handle_afs_exception(user_message=READ_ERROR)
    def read_path(self, path, arrow_flight_config, dataframe_type):
        descriptor = pyarrow.flight.FlightDescriptor.for_path(path)
        return self._get_dataset(
            descriptor,
            timeout=arrow_flight_config.get("timeout", self.timeout)
            if arrow_flight_config
            else self.timeout,
            dataframe_type=dataframe_type,
        )

    @_handle_afs_exception(user_message=WRITE_ERROR)
    @retry(
        wait_exponential_multiplier=1000,
        stop_max_attempt_number=3,
        # this retry only on FlightUnavailableError, which is only raised as long as
        # do_action call did not trigger any write on the server
        # Technically same as _should_retry, used for read operations but to avoid future mistakes
        # it is better to have separate retry_on_exception function with explicit limitations
        retry_on_exception=lambda e: isinstance(
            e, pyarrow._flight.FlightUnavailableError
        ),
    )
    def create_training_dataset(
        self, feature_view_obj, training_dataset_obj, query_obj, arrow_flight_config
    ):
        training_dataset = {}
        training_dataset["project_name"] = self._client._project_name
        training_dataset["fv_name"] = feature_view_obj.name
        training_dataset["fv_version"] = feature_view_obj.version
        training_dataset["tds_version"] = training_dataset_obj.version
        training_dataset["query"] = query_obj
        _logger.debug(f"Creating training dataset: {training_dataset}")
        try:
            training_dataset_encoded = json.dumps(training_dataset).encode("ascii")
            training_dataset_buf = pyarrow.py_buffer(training_dataset_encoded)
            action = pyarrow.flight.Action(
                "create-training-dataset", training_dataset_buf
            )
            timeout = (
                arrow_flight_config.get("timeout", self.timeout)
                if arrow_flight_config
                else self.timeout
            )
            options = pyarrow.flight.FlightCallOptions(
                timeout=timeout, headers=self._certificates_headers()
            )
            for result in self._connection.do_action(action, options):
                return result.body.to_pybytes()
        except pyarrow.lib.ArrowIOError as e:
            _logger.debug("Caught ArrowIOError in create_training_dataset: %s", str(e))
            _logger.exception(e)
            print("Error calling action:", e)

    def create_query_object(self, query, query_str, on_demand_fg_aliases=None):
        if on_demand_fg_aliases is None:
            on_demand_fg_aliases = []
        features = {}
        connectors = {}
        for fg in query.featuregroups:
            fg_name = _serialize_featuregroup_name(fg)
            fg_connector = _serialize_featuregroup_connector(
                fg, query, on_demand_fg_aliases
            )
            features[fg_name] = [feat.name for feat in fg.features]
            connectors[fg_name] = fg_connector
        filters = _serialize_filter_expression(query.filters, query)

        query = {
            "query_string": _translate_to_duckdb(query, query_str),
            "features": features,
            "filters": filters,
            "connectors": connectors,
        }
        return query

    def is_enabled(self):
        if self._disabled_for_session or not self._enabled_on_cluster:
            return False
        return True

    @property
    def timeout(self) -> Union[int, float]:
        """Timeout in seconds for Hopsworks Query Service do_get or do_action operations, not including the healthcheck."""
        return self._timeout

    @timeout.setter
    def timeout(self, value: Union[int, float]) -> None:
        self._timeout = value

    @property
    def health_check_timeout(self) -> Union[int, float]:
        """Timeout in seconds for the healthcheck operation."""
        return self._health_check_timeout

    @health_check_timeout.setter
    def health_check_timeout(self, value: Union[int, float]) -> None:
        self._health_check_timeout = value

    @property
    def host_url(self) -> Optional[str]:
        """URL of Hopsworks Query Service."""
        return self._host_url

    @host_url.setter
    def host_url(self, value: str) -> None:
        self._host_url = value

    @property
    def disabled_for_session(self) -> bool:
        """Whether the client is disabled for the current session."""
        return self._disabled_for_session

    @property
    def enabled_on_cluster(self) -> bool:
        """Whether the client is enabled on the cluster."""
        return self._enabled_on_cluster


def _serialize_featuregroup_connector(fg, query, on_demand_fg_aliases):
    connector = {}
    if isinstance(fg, feature_group.ExternalFeatureGroup):
        connector["time_travel_type"] = None
        connector["type"] = fg.storage_connector.type
        connector["options"] = fg.storage_connector.connector_options()
        connector["query"] = fg.query[:-1] if fg.query.endswith(";") else fg.query
        for on_demand_fg_alias in on_demand_fg_aliases:
            if on_demand_fg_alias.on_demand_feature_group.name == fg.name:
                connector["alias"] = on_demand_fg_alias.alias
                break
        if query._left_feature_group == fg:
            connector["filters"] = _serialize_filter_expression(
                query._filter, query, True
            )
        else:
            for join_obj in query._joins:
                if join_obj._query._left_feature_group == fg:
                    connector["filters"] = _serialize_filter_expression(
                        join_obj._query._filter, join_obj._query, True
                    )
    elif fg.time_travel_format == "DELTA":
        connector["time_travel_type"] = "delta"
        connector["type"] = fg.storage_connector.type
        connector["options"] = fg.storage_connector.connector_options()
        if fg.storage_connector.type == StorageConnector.S3:
            connector["options"]["path"] = fg.location
        connector["query"] = ""
        if query._left_feature_group == fg:
            connector["filters"] = _serialize_filter_expression(
                query._filter, query, True
            )
        else:
            for join_obj in query._joins:
                if join_obj._query._left_feature_group == fg:
                    connector["filters"] = _serialize_filter_expression(
                        join_obj._query._filter, join_obj._query, True
                    )
    else:
        connector["time_travel_type"] = "hudi"
    return connector


def _serialize_featuregroup_name(fg):
    return f"{fg._get_project_name()}.{fg.name}_{fg.version}"


def _serialize_filter_expression(filters, query, short_name=False):
    if filters is None:
        return None
    return _serialize_logic(filters, query, short_name)


def _serialize_logic(logic, query, short_name):
    return {
        "type": "logic",
        "logic_type": logic._type,
        "left_filter": _serialize_filter_or_logic(
            logic._left_f, logic._left_l, query, short_name
        ),
        "right_filter": _serialize_filter_or_logic(
            logic._right_f, logic._right_l, query, short_name
        ),
    }


def _serialize_filter_or_logic(filter, logic, query, short_name):
    if filter:
        return _serialize_filter(filter, query, short_name)
    elif logic:
        return _serialize_logic(logic, query, short_name)
    else:
        return None


def _serialize_filter(filter, query, short_name):
    if isinstance(filter._value, datetime.datetime):
        filter_value = filter._value.strftime("%Y-%m-%d %H:%M:%S")
    else:
        filter_value = filter._value

    return {
        "type": "filter",
        "condition": filter._condition,
        "value": filter_value,
        "feature": _serialize_feature_name(filter._feature, query, short_name),
    }


def _serialize_feature_name(feature, query, short_name):
    if short_name:
        return feature.name
    fg = query._get_featuregroup_by_feature(feature)
    fg_name = _serialize_featuregroup_name(fg)
    return f"{fg_name}.{feature.name}"


def _translate_to_duckdb(query, query_str):
    translated = query_str
    for fg in query.featuregroups:
        translated = translated.replace(
            f"`{fg.feature_store_name}`.`",
            f"`{fg._get_project_name()}.",
        )
    return translated.replace("`", '"')


def supports(featuregroups):
    if len(featuregroups) > sum(
        1
        for fg in featuregroups
        if isinstance(
            fg,
            (feature_group.FeatureGroup, feature_group.ExternalFeatureGroup),
        )
    ):
        # Contains unsupported feature group types such as a spine group
        return False

    for fg in filter(
        lambda fg: isinstance(fg, feature_group.ExternalFeatureGroup), featuregroups
    ):
        if (
            fg.storage_connector.type
            not in ArrowFlightClient.SUPPORTED_EXTERNAL_CONNECTORS
        ):
            return False
    return True



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This implementation has been borrowed from kserve/kserve repository
# https://github.com/kserve/kserve/blob/release-0.11/python/kserve/kserve/utils/numpy_codec.py

import numpy as np


def to_np_dtype(dtype):
    dtype_map = {
        "BOOL": bool,
        "INT8": np.int8,
        "INT16": np.int16,
        "INT32": np.int32,
        "INT64": np.int64,
        "UINT8": np.uint8,
        "UINT16": np.uint16,
        "UINT32": np.uint32,
        "UINT64": np.uint64,
        "FP16": np.float16,
        "FP32": np.float32,
        "FP64": np.float64,
        "BYTES": np.object_,
    }
    return dtype_map.get(dtype, None)


def from_np_dtype(np_dtype):
    if np_dtype is bool:
        return "BOOL"
    elif np_dtype == np.int8:
        return "INT8"
    elif np_dtype == np.int16:
        return "INT16"
    elif np_dtype == np.int32:
        return "INT32"
    elif np_dtype == np.int64:
        return "INT64"
    elif np_dtype == np.uint8:
        return "UINT8"
    elif np_dtype == np.uint16:
        return "UINT16"
    elif np_dtype == np.uint32:
        return "UINT32"
    elif np_dtype == np.uint64:
        return "UINT64"
    elif np_dtype == np.float16:
        return "FP16"
    elif np_dtype == np.float32:
        return "FP32"
    elif np_dtype == np.float64:
        return "FP64"
    elif np_dtype == np.object_ or np_dtype.type == np.bytes_:
        return "BYTES"
    return None



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from datetime import datetime

import humps
from hopsworks_common import util


class DeployableComponentLogs:
    """Server logs of a deployable component (predictor or transformer).

    # Arguments
        name: Deployment instance name.
        content: actual logs
    # Returns
        `DeployableComponentLogs`. Server logs of a deployable component
    """

    def __init__(self, instance_name: str, content: str, **kwargs):
        self._instance_name = instance_name
        self._content = content
        self._created_at = datetime.now()

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        if len(json_decamelized) == 0:
            return []
        return [cls.from_json(logs) for logs in json_decamelized]

    @classmethod
    def from_json(cls, json_decamelized):
        return DeployableComponentLogs(*cls.extract_fields_from_json(json_decamelized))

    @classmethod
    def extract_fields_from_json(cls, json_decamelized):
        instance_name = util.extract_field_from_json(json_decamelized, "instance_name")
        content = util.extract_field_from_json(json_decamelized, "content")
        return instance_name, content

    def to_dict(self):
        return {"instance_name": self._instance_name, "content": self._content}

    @property
    def instance_name(self):
        """Name of the deployment instance containing these server logs."""
        return self._instance_name

    @property
    def content(self):
        """Content of the server logs of the current deployment instance."""
        return self._content

    @property
    def created_at(self):
        """Datetime when the current server logs chunk was retrieved."""
        return self._created_at

    @property
    def component(self):
        """Component of the deployment containing these server logs."""
        return self._component

    @component.setter
    def component(self, component: str):
        self._component = component

    @property
    def tail(self):
        """Number of lines of server logs."""
        return self._tail

    @tail.setter
    def tail(self, tail: int):
        self._tail = tail

    def __repr__(self):
        return f"DeployableComponentLogs(instance_name: {self._instance_name!r}, date: {self._created_at!r}) \n{self._content!s}"



#
#   Copyright 2023 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

from typing import Any, Dict

from furl import furl
from hopsworks_common import client, usage
from hopsworks_common.client.exceptions import FeatureStoreException
from hopsworks_common.core.variable_api import VariableApi


class OPENSEARCH_CONFIG:
    ELASTIC_ENDPOINT_ENV_VAR = "ELASTIC_ENDPOINT"
    SSL_CONFIG = "es.net.ssl"
    NODES_WAN_ONLY = "es.nodes.wan.only"
    NODES = "es.nodes"
    SSL_KEYSTORE_LOCATION = "es.net.ssl.keystore.location"
    SSL_KEYSTORE_PASSWORD = "es.net.ssl.keystore.pass"
    SSL_TRUSTSTORE_LOCATION = "es.net.ssl.truststore.location"
    SSL_TRUSTSTORE_PASSWORD = "es.net.ssl.truststore.pass"
    HTTP_AUTHORIZATION = "es.net.http.header.Authorization"
    INDEX = "es.resource"
    HOSTS = "hosts"
    HTTP_COMPRESS = "http_compress"
    HEADERS = "headers"
    USE_SSL = "use_ssl"
    VERIFY_CERTS = "verify_certs"
    SSL_ASSERT_HOSTNAME = "ssl_assert_hostname"
    CA_CERTS = "ca_certs"


class OpenSearchApi:
    def __init__(self) -> None:
        self._variable_api: VariableApi = VariableApi()

    def _get_opensearch_url(self) -> str:
        if client._is_external():
            external_domain = self._variable_api.get_loadbalancer_external_domain(
                "opensearch"
            )
            return f"https://{external_domain}:9200"
        else:
            service_discovery_domain = self._variable_api.get_service_discovery_domain()
            if service_discovery_domain == "":
                raise FeatureStoreException(
                    "Client could not locate service_discovery_domain "
                    "in Hopsworks cluster configuration or variable is empty."
                )
            return f"https://rest.elastic.service.{service_discovery_domain}:9200"

    @usage.method_logger
    def get_project_index(self, index: str) -> str:
        """
        This helper method prefixes the supplied index name with the project name to avoid index name clashes.

        # Arguments
            index: the opensearch index to interact with.

        # Returns
            `str`: A valid opensearch index name.
        """
        _client = client.get_instance()
        return (_client._project_name + "_" + index).lower()

    @usage.method_logger
    def get_default_py_config(self) -> Dict[str, Any]:
        """
        Get the required opensearch configuration to setup a connection using the *opensearch-py* library.

        ```python

        import hopsworks
        from opensearchpy import OpenSearch

        project = hopsworks.login()

        opensearch_api = project.get_opensearch_api()

        client = OpenSearch(**opensearch_api.get_default_py_config())

        ```

        # Returns
            `dict`: A dictionary with required configuration.
        """
        url = furl(self._get_opensearch_url())
        return {
            OPENSEARCH_CONFIG.HOSTS: [{"host": url.host, "port": url.port}],
            OPENSEARCH_CONFIG.HTTP_COMPRESS: False,
            OPENSEARCH_CONFIG.HEADERS: {
                "Authorization": self._get_authorization_token()
            },
            OPENSEARCH_CONFIG.USE_SSL: True,
            OPENSEARCH_CONFIG.VERIFY_CERTS: True,
            OPENSEARCH_CONFIG.SSL_ASSERT_HOSTNAME: False,
            OPENSEARCH_CONFIG.CA_CERTS: client.get_instance()._get_ca_chain_path(),
        }

    def _get_authorization_token(self) -> str:
        """Get opensearch jwt token.

        # Returns
            `str`: OpenSearch jwt token
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """

        _client = client.get_instance()
        path_params = ["elastic", "jwt", _client._project_id]

        headers = {"content-type": "application/json"}
        return _client._send_request("GET", path_params, headers=headers)["token"]



#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#


import logging

import pandas as pd
import pytest
from hopsworks_common import version
from hsfs.client.exceptions import FeatureStoreException
from hsfs.hopsworks_udf import udf
from hsfs.transformation_function import TransformationFunction, TransformationType
from packaging.version import Version


class TestTransformationFunction:
    def test_from_response_json_one_argument_no_statistics(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["transformation_function"][
            "get_one_argument_no_statistics_function"
        ]["response"]
        json["transformation_type"] = TransformationType.MODEL_DEPENDENT
        # Act
        tf = TransformationFunction.from_response_json(json)

        # Assert
        assert tf.id == 1
        assert tf._featurestore_id == 11
        assert tf.version == 2
        assert tf.hopsworks_udf.function_name == "add_one_fs"
        assert tf.hopsworks_udf.return_types == ["double"]
        assert not tf.hopsworks_udf.statistics_required
        assert tf.hopsworks_udf.transformation_features == ["col1"]
        assert tf.hopsworks_udf.statistics_features == []
        assert tf.hopsworks_udf._statistics_argument_names == []
        assert (
            tf.hopsworks_udf._function_source
            == "\n@udf(float)\ndef add_one_fs(data1 : pd.Series):\n    return data1 + 1\n"
        )

    def test_from_response_json_one_argument_with_statistics(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["transformation_function"][
            "get_one_argument_with_statistics_function"
        ]["response"]
        json["transformation_type"] = TransformationType.MODEL_DEPENDENT

        # Act
        tf = TransformationFunction.from_response_json(json)

        # Assert
        assert tf.id == 1
        assert tf._featurestore_id == 11
        assert tf.version == 2
        assert tf.hopsworks_udf.function_name == "add_mean_fs"
        assert tf.hopsworks_udf.return_types == ["double"]
        assert tf.hopsworks_udf.statistics_required
        assert tf.hopsworks_udf.transformation_features == ["data"]
        assert tf.hopsworks_udf.statistics_features == ["data"]
        assert tf.hopsworks_udf._statistics_argument_names == ["data1"]
        assert (
            tf.hopsworks_udf._function_source
            == "\n@udf(float)\ndef add_mean_fs(data1 : pd.Series, statistics=stats):\n    return data1 + statistics.data1.mean\n"
        )

    def test_from_response_json_multiple_argument_with_statistics(
        self, backend_fixtures
    ):
        # Arrange
        json = backend_fixtures["transformation_function"][
            "get_multiple_argument_with_statistics_function"
        ]["response"]
        json["transformation_type"] = TransformationType.MODEL_DEPENDENT

        # Act
        tf = TransformationFunction.from_response_json(json)

        # Assert
        assert tf.id == 1
        assert tf._featurestore_id == 11
        assert tf.version == 2
        assert tf.hopsworks_udf.function_name == "test_func"
        assert tf.hopsworks_udf.return_types == ["string"]
        assert tf.hopsworks_udf.statistics_required
        assert tf.hopsworks_udf.transformation_features == [
            "feature1",
            "feature2",
            "feature3",
        ]
        assert tf.hopsworks_udf.statistics_features == ["feature1", "feature2"]
        assert tf.hopsworks_udf._statistics_argument_names == ["data1", "data2"]
        assert (
            tf.hopsworks_udf._function_source
            == "\n@udf(str)\ndef test_func(data1 : pd.Series, data2, data3, statistics=stats):\n    return data1 + statistics.data1.mean\n"
        )

    def test_from_response_json_multiple_return_type_functions(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["transformation_function"][
            "get_multiple_return_type_functions"
        ]["response"]
        json["transformation_type"] = TransformationType.MODEL_DEPENDENT

        # Act
        tf = TransformationFunction.from_response_json(json)

        # Assert
        assert tf.id == 1
        assert tf._featurestore_id == 11
        assert tf.version == 2
        assert tf.hopsworks_udf.function_name == "test_func"
        assert tf.hopsworks_udf.return_types == ["string", "double"]
        assert tf.hopsworks_udf.statistics_required
        assert tf.hopsworks_udf.transformation_features == [
            "feature1",
            "feature2",
            "feature3",
        ]
        assert tf.hopsworks_udf.statistics_features == ["feature1", "feature2"]
        assert tf.hopsworks_udf._statistics_argument_names == ["data1", "data2"]
        assert (
            tf.hopsworks_udf._function_source
            == "\n@udf(str, float)\ndef test_func(data1 : pd.Series, data2, data3, statistics=stats):\n    return pd.DataFrame('col1': ['a', 'b'], 'col2':[1,2])\n"
        )

    def test_from_response_json_list_empty(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["transformation_function"]["get_list_empty"]["response"]

        # Act
        tf_list = TransformationFunction.from_response_json(json)

        # Assert
        assert len(tf_list) == 0

    def test_from_response_json_list(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["transformation_function"]["get_list"]["response"]
        for response_json in json["items"]:
            response_json["transformation_type"] = TransformationType.MODEL_DEPENDENT

        # Act
        tf_list = TransformationFunction.from_response_json(json)

        # Assert
        assert len(tf_list) == 2
        tf = tf_list[0]
        assert tf.id == 1
        assert tf._featurestore_id == 11
        assert tf.version == 2
        assert tf.hopsworks_udf.function_name == "add_mean_fs"
        assert tf.hopsworks_udf.return_types == ["double"]
        assert tf.hopsworks_udf.statistics_required
        assert tf.hopsworks_udf.transformation_features == ["data"]
        assert tf.hopsworks_udf.statistics_features == ["data"]
        assert tf.hopsworks_udf._statistics_argument_names == ["data1"]
        assert (
            tf.hopsworks_udf._function_source
            == "\n@udf(float)\ndef add_mean_fs(data1 : pd.Series, statistics=stats):\n    return data1 + statistics.data1.mean\n"
        )

        tf = tf_list[1]
        assert tf.id == 2
        assert tf._featurestore_id == 11
        assert tf.version == 1
        assert tf.hopsworks_udf.function_name == "add_one_fs"
        assert tf.hopsworks_udf.return_types == ["double"]
        assert not tf.hopsworks_udf.statistics_required
        assert tf.hopsworks_udf.transformation_features == ["col1"]
        assert tf.hopsworks_udf.statistics_features == []
        assert tf.hopsworks_udf._statistics_argument_names == []
        assert (
            tf.hopsworks_udf._function_source
            == "\n@udf(float)\ndef add_one_fs(data1 : pd.Series):\n    return data1 + 1\n"
        )

    def test_from_response_json_list_one_argument(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["transformation_function"]["get_list_one_argument"][
            "response"
        ]
        for response_json in json["items"]:
            response_json["transformation_type"] = TransformationType.MODEL_DEPENDENT

        # Act
        tf = TransformationFunction.from_response_json(json)

        # Assert
        assert not isinstance(tf, list)
        assert tf.id == 1
        assert tf._featurestore_id == 11
        assert tf.version == 2
        assert tf.hopsworks_udf.function_name == "add_mean_fs"
        assert tf.hopsworks_udf.return_types == ["double"]
        assert tf.hopsworks_udf.statistics_required
        assert tf.hopsworks_udf.transformation_features == ["data"]
        assert tf.hopsworks_udf.statistics_features == ["data"]
        assert tf.hopsworks_udf._statistics_argument_names == ["data1"]
        assert (
            tf.hopsworks_udf._function_source
            == "\n@udf(float)\ndef add_mean_fs(data1 : pd.Series, statistics=stats):\n    return data1 + statistics.data1.mean\n"
        )

    def test_transformation_function_definition_no_hopworks_udf(self):
        def test(col1):
            return col1 + 1

        with pytest.raises(FeatureStoreException) as exception:
            TransformationFunction(
                featurestore_id=10,
                hopsworks_udf=test,
                transformation_type=TransformationType.MODEL_DEPENDENT,
            )

        assert (
            str(exception.value)
            == "Please use the hopsworks_udf decorator when defining transformation functions."
        )

    def test_transformation_function_definition_with_hopworks_udf(self, mocker):
        mocked_connection = mocker.MagicMock()
        mocked_connection.backend_version = (
            version.__version__
        )  # Mocking backend version to be the same as the current version
        mocked_connection = mocker.patch(
            "hopsworks_common.client.get_connection", return_value=mocked_connection
        )

        @udf(int)
        def test2(col1):
            return col1 + 1

        tf = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=test2,
            transformation_type=TransformationType.MODEL_DEPENDENT,
        )

        # Creating dict representation of udf.
        udf_json = test2.to_dict()
        # Adding output column names to dict for testing since it would be generated when UDF is accessed out the transformation function.
        udf_json["outputColumnNames"] = ["test2_col1_"]

        assert tf.hopsworks_udf.to_dict() == udf_json

    def test_generate_output_column_names_one_argument_one_output_type_mdt(self):
        @udf(int)
        def test_func(col1):
            return col1 + 1

        mdt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=test_func,
            transformation_type=TransformationType.MODEL_DEPENDENT,
        )
        assert mdt._get_output_column_names() == ["test_func_col1_"]

    def test_generate_output_column_names_one_argument_one_output_type_odt(self):
        @udf(int)
        def test_func(col1):
            return col1 + 1

        odt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=test_func,
            transformation_type=TransformationType.ON_DEMAND,
        )
        assert odt._get_output_column_names() == ["test_func"]

    def test_generate_output_column_names_one_argument_one_output_type_prefix_mdt(self):
        @udf(int)
        def test_func(col1):
            return col1 + 1

        test_func._feature_name_prefix = "prefix_"

        mdt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=test_func,
            transformation_type=TransformationType.MODEL_DEPENDENT,
        )
        assert mdt._get_output_column_names() == ["test_func_prefix_col1_"]

    def test_generate_output_column_names_one_argument_one_output_type_prefix_odt(self):
        @udf(int)
        def test_func(col1):
            return col1 + 1

        test_func._feature_name_prefix = "prefix_"

        odt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=test_func,
            transformation_type=TransformationType.ON_DEMAND,
        )
        assert odt._get_output_column_names() == ["test_func"]
        assert odt.output_column_names == ["prefix_test_func"]

    def test_generate_output_column_names_multiple_argument_one_output_type_mdt(self):
        @udf(int)
        def test_func(col1, col2, col3):
            return col1 + 1

        mdt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=test_func,
            transformation_type=TransformationType.MODEL_DEPENDENT,
        )
        assert mdt._get_output_column_names() == ["test_func_col1_col2_col3_"]

    def test_generate_output_column_names_multiple_argument_one_output_type_odt(self):
        @udf(int)
        def test_func(col1, col2, col3):
            return col1 + 1

        odt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=test_func,
            transformation_type=TransformationType.ON_DEMAND,
        )
        assert odt._get_output_column_names() == ["test_func"]

    def test_generate_output_column_names_multiple_argument_one_output_type_prefix_mdt(
        self,
    ):
        @udf(int)
        def test_func(col1, col2, col3):
            return col1 + 1

        test_func._feature_name_prefix = "prefix_"

        mdt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=test_func,
            transformation_type=TransformationType.MODEL_DEPENDENT,
        )
        assert mdt._get_output_column_names() == [
            "test_func_prefix_col1_prefix_col2_prefix_col3_"
        ]
        assert mdt.output_column_names == [
            "prefix_test_func_prefix_col1_prefix_col2_prefix_col3_"
        ]

    def test_generate_output_column_names_multiple_argument_one_output_type_prefix_odt(
        self,
    ):
        @udf(int)
        def test_func(col1, col2, col3):
            return col1 + 1

        test_func._feature_name_prefix = "prefix_"

        odt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=test_func,
            transformation_type=TransformationType.ON_DEMAND,
        )
        assert odt._get_output_column_names() == ["test_func"]
        assert odt.output_column_names == ["prefix_test_func"]

    def test_generate_output_column_names_single_argument_multiple_output_type_mdt(
        self,
    ):
        @udf([int, float, int])
        def test_func(col1):
            return pd.DataFrame(
                {"col1": [col1 + 1], "col2": [col1 + 1], "col3": [col1 + 1]}
            )

        mdt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=test_func,
            transformation_type=TransformationType.MODEL_DEPENDENT,
        )
        assert mdt._get_output_column_names() == [
            "test_func_col1_0",
            "test_func_col1_1",
            "test_func_col1_2",
        ]

    def test_generate_output_column_names_single_argument_multiple_output_type_odt(
        self,
    ):
        @udf([int, float, int])
        def test_func(col1):
            return pd.DataFrame(
                {"col1": [col1 + 1], "col2": [col1 + 1], "col3": [col1 + 1]}
            )

        odt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=test_func,
            transformation_type=TransformationType.ON_DEMAND,
        )
        assert odt._get_output_column_names() == [
            "test_func_0",
            "test_func_1",
            "test_func_2",
        ]
        assert odt.output_column_names == ["test_func_0", "test_func_1", "test_func_2"]

    def test_generate_output_column_names_single_argument_multiple_output_type_prefix_mdt(
        self,
    ):
        @udf([int, float, int])
        def test_func(col1):
            return pd.DataFrame(
                {"col1": [col1 + 1], "col2": [col1 + 1], "col3": [col1 + 1]}
            )

        test_func._feature_name_prefix = "prefix_"

        mdt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=test_func,
            transformation_type=TransformationType.MODEL_DEPENDENT,
        )
        assert mdt._get_output_column_names() == [
            "test_func_prefix_col1_0",
            "test_func_prefix_col1_1",
            "test_func_prefix_col1_2",
        ]
        assert mdt.output_column_names == [
            "prefix_test_func_prefix_col1_0",
            "prefix_test_func_prefix_col1_1",
            "prefix_test_func_prefix_col1_2",
        ]

    def test_generate_output_column_names_single_argument_multiple_output_type_prefix_odt(
        self,
    ):
        @udf([int, float, int])
        def test_func(col1):
            return pd.DataFrame(
                {"col1": [col1 + 1], "col2": [col1 + 1], "col3": [col1 + 1]}
            )

        test_func._feature_name_prefix = "prefix_"

        odt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=test_func,
            transformation_type=TransformationType.ON_DEMAND,
        )

        assert odt._get_output_column_names() == [
            "test_func_0",
            "test_func_1",
            "test_func_2",
        ]
        assert odt.output_column_names == [
            "prefix_test_func_0",
            "prefix_test_func_1",
            "prefix_test_func_2",
        ]

    def test_generate_output_column_names_multiple_argument_multiple_output_type_mdt(
        self,
    ):
        @udf([int, float, int])
        def test_func(col1, col2, col3):
            return pd.DataFrame(
                {"col1": [col1 + 1], "col2": [col2 + 1], "col3": [col3 + 1]}
            )

        mdt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=test_func,
            transformation_type=TransformationType.MODEL_DEPENDENT,
        )
        assert mdt._get_output_column_names() == [
            "test_func_col1_col2_col3_0",
            "test_func_col1_col2_col3_1",
            "test_func_col1_col2_col3_2",
        ]
        assert mdt.output_column_names == [
            "test_func_col1_col2_col3_0",
            "test_func_col1_col2_col3_1",
            "test_func_col1_col2_col3_2",
        ]

    def test_generate_output_column_names_multiple_argument_multiple_output_type_odt(
        self,
    ):
        @udf([int, float, int])
        def test_func(col1, col2, col3):
            return pd.DataFrame(
                {"col1": [col1 + 1], "col2": [col2 + 1], "col3": [col3 + 1]}
            )

        odt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=test_func,
            transformation_type=TransformationType.ON_DEMAND,
        )

        assert odt._get_output_column_names() == [
            "test_func_0",
            "test_func_1",
            "test_func_2",
        ]
        assert odt.output_column_names == ["test_func_0", "test_func_1", "test_func_2"]

    def test_generate_output_column_names_multiple_argument_multiple_output_type_prefix_mdt(
        self,
    ):
        @udf([int, float, int])
        def test_func(col1, col2, col3):
            return pd.DataFrame(
                {"col1": [col1 + 1], "col2": [col2 + 1], "col3": [col3 + 1]}
            )

        test_func._feature_name_prefix = "prefix_"

        mdt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=test_func,
            transformation_type=TransformationType.MODEL_DEPENDENT,
        )
        assert mdt._get_output_column_names() == [
            "test_func_prefix_col1_prefix_col2_prefix_col3_0",
            "test_func_prefix_col1_prefix_col2_prefix_col3_1",
            "test_func_prefix_col1_prefix_col2_prefix_col3_2",
        ]
        assert mdt.output_column_names == [
            "prefix_test_func_prefix_col1_prefix_col2_prefix_col3_0",
            "prefix_test_func_prefix_col1_prefix_col2_prefix_col3_1",
            "prefix_test_func_prefix_col1_prefix_col2_prefix_col3_2",
        ]

    def test_generate_output_column_names_multiple_argument_multiple_output_type_prefix_odt(
        self,
    ):
        @udf([int, float, int])
        def test_func(col1, col2, col3):
            return pd.DataFrame(
                {"col1": [col1 + 1], "col2": [col2 + 1], "col3": [col3 + 1]}
            )

        test_func._feature_name_prefix = "prefix_"

        odt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=test_func,
            transformation_type=TransformationType.ON_DEMAND,
        )

        assert odt._get_output_column_names() == [
            "test_func_0",
            "test_func_1",
            "test_func_2",
        ]
        assert odt.output_column_names == [
            "prefix_test_func_0",
            "prefix_test_func_1",
            "prefix_test_func_2",
        ]

    def test_validate_udf_type_on_demand_statistics(self):
        from hsfs.transformation_statistics import TransformationStatistics

        stats = TransformationStatistics("col1")

        @udf(int)
        def test_func(col1, statistics=stats):
            return col1 + statistics.col1.mean

        with pytest.raises(FeatureStoreException) as exe:
            TransformationFunction(
                featurestore_id=10,
                hopsworks_udf=test_func,
                transformation_type=TransformationType.ON_DEMAND,
            )

        assert (
            str(exe.value)
            == "On-Demand Transformation functions cannot use statistics, please remove statistics parameters from the functions"
        )

    def test_alias_one_output_mdt(self):
        @udf(int)
        def add_one(feature):
            return feature + 1

        mdt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=add_one,
            transformation_type=TransformationType.MODEL_DEPENDENT,
        )

        mdt = mdt.alias("feature_plus_one_mdt")

        assert mdt.output_column_names == ["feature_plus_one_mdt"]

    def test_alias_one_output_odt(self):
        @udf(int)
        def add_one(feature):
            return feature + 1

        odt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=add_one,
            transformation_type=TransformationType.ON_DEMAND,
        )

        odt = odt.alias("feature_plus_one_odt")

        assert odt.output_column_names == ["feature_plus_one_odt"]

    def test_alias_one_output_list_mdt(self):
        @udf(int)
        def add_one(feature):
            return feature + 1

        mdt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=add_one,
            transformation_type=TransformationType.MODEL_DEPENDENT,
        )

        mdt = mdt.alias(["feature_plus_one_mdt"])

        assert mdt.output_column_names == ["feature_plus_one_mdt"]

    def test_alias_one_output_list_odt(self):
        @udf(int)
        def add_one(feature):
            return feature + 1

        odt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=add_one,
            transformation_type=TransformationType.ON_DEMAND,
        )

        odt = odt.alias(["feature_plus_one_odt"])

        assert odt.output_column_names == ["feature_plus_one_odt"]

    def test_alias_multiple_output_mdt(self):
        @udf([int, int])
        def add_and_sub(feature):
            return feature + 1, feature - 1

        mdt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=add_and_sub,
            transformation_type=TransformationType.MODEL_DEPENDENT,
        )

        mdt = mdt.alias("feature_plus_one_mdt", "feature_minus_one_mdt")

        assert mdt.output_column_names == [
            "feature_plus_one_mdt",
            "feature_minus_one_mdt",
        ]

    def test_alias_multiple_output_list_mdt(self):
        @udf([int, int])
        def add_and_sub(feature):
            return feature + 1, feature - 1

        mdt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=add_and_sub,
            transformation_type=TransformationType.MODEL_DEPENDENT,
        )

        mdt = mdt.alias(["feature_plus_one_mdt", "feature_minus_one_mdt"])

        assert mdt.output_column_names == [
            "feature_plus_one_mdt",
            "feature_minus_one_mdt",
        ]

    def test_alias_invalid_number_column_names_mdt(self):
        @udf([int, int])
        def add_and_sub(feature):
            return feature + 1, feature - 1

        mdt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=add_and_sub,
            transformation_type=TransformationType.MODEL_DEPENDENT,
        )

        with pytest.raises(FeatureStoreException) as exp:
            mdt.alias(["feature_plus_one", "feature_minus_one", "invalid_col"])

        assert (
            str(exp.value)
            == "The number of output feature names provided does not match the number of features returned by the transformation function 'add_and_sub(feature)'. Pease provide exactly 2 feature name(s) to match the output."
        )

    def test_alias_invalid_number_column_names_odt(self):
        @udf(int)
        def add_one(feature):
            return feature + 1

        odt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=add_one,
            transformation_type=TransformationType.ON_DEMAND,
        )

        with pytest.raises(FeatureStoreException) as exp:
            odt.alias(["feature_plus_one", "feature_minus_one", "invalid_col"])

        assert (
            str(exp.value)
            == "The number of output feature names provided does not match the number of features returned by the transformation function 'add_one(feature)'. Pease provide exactly 1 feature name(s) to match the output."
        )

    def test_alias_invalid_type_mdt(self):
        @udf([int])
        def add_one(feature):
            return feature + 1

        mdt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=add_one,
            transformation_type=TransformationType.MODEL_DEPENDENT,
        )

        with pytest.raises(FeatureStoreException) as exp:
            mdt.alias({"name": "col1"})

        assert (
            str(exp.value)
            == "Invalid output feature names provided for the transformation function 'add_one(feature)'. Please ensure all arguments are strings."
        )

    def test_alias_invalid_type_odt(self):
        @udf([int])
        def add_one(feature):
            return feature + 1

        odt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=add_one,
            transformation_type=TransformationType.ON_DEMAND,
        )

        with pytest.raises(FeatureStoreException) as exp:
            odt.alias({"name": "col1"})

        assert (
            str(exp.value)
            == "Invalid output feature names provided for the transformation function 'add_one(feature)'. Please ensure all arguments are strings."
        )

    def test_alias_duplicates_mdt(self):
        @udf([int, int])
        def add_and_sub(feature):
            return feature + 1, feature - 1

        mdt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=add_and_sub,
            transformation_type=TransformationType.MODEL_DEPENDENT,
        )

        with pytest.raises(FeatureStoreException) as exp:
            mdt.alias("feature_plus_one", "feature_plus_one")

        assert (
            str(exp.value)
            == "Duplicate output feature names provided for the transformation function 'add_and_sub(feature)'. Please ensure all arguments names are unique."
        )

    def test_call_and_alias_mdt(self):
        @udf(int)
        def add_one(feature):
            return feature + 1

        mdt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=add_one,
            transformation_type=TransformationType.MODEL_DEPENDENT,
        )

        mdt = mdt("feature2_mdt").alias(["feature_plus_one_mdt"])

        assert mdt.output_column_names == ["feature_plus_one_mdt"]
        assert mdt.hopsworks_udf.transformation_features == ["feature2_mdt"]

    def test_call_and_alias_odt(self):
        @udf(int)
        def add_one(feature):
            return feature + 1

        odt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=add_one,
            transformation_type=TransformationType.ON_DEMAND,
        )

        odt = odt("feature2_odt").alias(["feature_plus_one_odt"])

        assert odt.output_column_names == ["feature_plus_one_odt"]
        assert odt.hopsworks_udf.transformation_features == ["feature2_odt"]

    def test_alias_invalid_length_mdt(self):
        @udf(int)
        def add_one(feature):
            return feature + 1

        mdt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=add_one,
            transformation_type=TransformationType.MODEL_DEPENDENT,
        )

        with pytest.raises(FeatureStoreException) as exp:
            mdt.alias(["invalid" * 10])

        assert (
            str(exp.value)
            == "Invalid output feature names specified for the transformation function 'add_one(feature)'. Please provide names shorter than 63 characters."
        )

    def test_alias_invalid_length_odt(self):
        @udf(int)
        def add_one(feature):
            return feature + 1

        odt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=add_one,
            transformation_type=TransformationType.ON_DEMAND,
        )

        with pytest.raises(FeatureStoreException) as exp:
            odt.alias(["invalid" * 10])

        assert (
            str(exp.value)
            == "Invalid output feature names specified for the transformation function 'add_one(feature)'. Please provide names shorter than 63 characters."
        )

    def test_generated_output_col_name_invalid_mdt(self, caplog):
        @udf(int)
        def test(
            long_feature_name1,
            long_feature_name2,
            long_feature_name3,
            long_feature_name4,
            long_feature_name5,
            long_feature_name6,
            long_feature_name7,
        ):
            return long_feature_name1

        mdt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=test,
            transformation_type=TransformationType.MODEL_DEPENDENT,
        )

        with caplog.at_level(logging.WARNING):
            mdt._get_output_column_names()

        assert (
            "The default output feature names generated by the transformation function test(long_feature_name1, long_feature_name2, long_feature_name3, long_feature_name4, long_feature_name5, long_feature_name6, long_feature_name7) exceed the maximum allowed length of 63 characters. Default names have been truncated to fit within the size limit. To avoid this, consider using the alias function to explicitly specify output column names."
            in caplog.text
        )

        assert mdt.output_column_names == [
            "test_long_feature_name1_long_feature_name2_long_feature_name3_l"
        ]

    def test_generated_output_col_name_invalid_mdt_multiple_returns(self, caplog):
        @udf([int, int, int])
        def test(
            long_feature_name1,
            long_feature_name2,
            long_feature_name3,
            long_feature_name4,
            long_feature_name5,
            long_feature_name6,
            long_feature_name7,
        ):
            return long_feature_name1, long_feature_name2, long_feature_name3

        mdt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=test,
            transformation_type=TransformationType.MODEL_DEPENDENT,
        )

        with caplog.at_level(logging.WARNING):
            mdt._get_output_column_names()

        assert (
            "The default output feature names generated by the transformation function test(long_feature_name1, long_feature_name2, long_feature_name3, long_feature_name4, long_feature_name5, long_feature_name6, long_feature_name7) exceed the maximum allowed length of 63 characters. Default names have been truncated to fit within the size limit. To avoid this, consider using the alias function to explicitly specify output column names."
            in caplog.text
        )

        assert mdt.output_column_names == [
            "test_long_feature_name1_long_feature_name2_long_feature_nam_0",
            "test_long_feature_name1_long_feature_name2_long_feature_nam_1",
            "test_long_feature_name1_long_feature_name2_long_feature_nam_2",
        ]

    def test_generate_output_col_name_invalid_odt(self, caplog):
        @udf(int)
        def really_long_function_name_that_exceed_63_characters_causing_invalid_name_for_on_demand_features(
            features,
        ):
            return features

        odt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=really_long_function_name_that_exceed_63_characters_causing_invalid_name_for_on_demand_features,
            transformation_type=TransformationType.ON_DEMAND,
        )

        with caplog.at_level(logging.WARNING):
            odt._get_output_column_names()

        assert (
            "The default output feature names generated by the transformation function really_long_function_name_that_exceed_63_characters_causing_invalid_name_for_on_demand_features(features) exceed the maximum allowed length of 63 characters. Default names have been truncated to fit within the size limit. To avoid this, consider using the alias function to explicitly specify output column names."
            in caplog.text
        )

        assert odt.hopsworks_udf.output_column_names == [
            "really_long_function_name_that_exceed_63_characters_causing_inv"
        ]

    def test_generate_output_col_name_invalid_mdt(self, caplog):
        @udf(int)
        def really_long_function_name_that_exceed_63_characters_causing_invalid_name_for_on_demand_features(
            features,
        ):
            return features

        mdt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=really_long_function_name_that_exceed_63_characters_causing_invalid_name_for_on_demand_features,
            transformation_type=TransformationType.MODEL_DEPENDENT,
        )

        with caplog.at_level(logging.WARNING):
            mdt._get_output_column_names()

        assert (
            "The default output feature names generated by the transformation function really_long_function_name_that_exceed_63_characters_causing_invalid_name_for_on_demand_features(features) exceed the maximum allowed length of 63 characters. Default names have been truncated to fit within the size limit. To avoid this, consider using the alias function to explicitly specify output column names."
            in caplog.text
        )

        # Asserting that the output column names generate are stripped to the required length
        assert mdt.hopsworks_udf.output_column_names == [
            "really_long_function_name_that_exceed_63_characters_causing_inv"
        ]

    @pytest.mark.skipif(
        Version("4.1.6") >= Version(version.__version__),
        reason="Requires Hopsworks 4.1.7 or higher to be working.",
    )
    def test_equality_mdt(self, mocker):
        mocked_connection = mocker.MagicMock()
        mocked_connection.backend_version = (
            version.__version__
        )  # Mocking backend version to be the same as the current version
        mocked_connection = mocker.patch(
            "hopsworks_common.client.get_connection", return_value=mocked_connection
        )

        @udf([int])
        def add_one(feature):
            return feature + 1

        mdt1 = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=add_one,
            transformation_type=TransformationType.MODEL_DEPENDENT,
        )

        mdt2 = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=add_one,
            transformation_type=TransformationType.MODEL_DEPENDENT,
        )

        assert mdt1 == mdt2

    @pytest.mark.skipif(
        Version("4.1.6") >= Version(version.__version__),
        reason="Requires Hopsworks 4.1.7 or higher to be working.",
    )
    def test_equality_odt(self, mocker):
        mocked_connection = mocker.MagicMock()
        mocked_connection.backend_version = (
            version.__version__
        )  # Mocking backend version to be the same as the current version
        mocked_connection = mocker.patch(
            "hopsworks_common.client.get_connection", return_value=mocked_connection
        )

        @udf([int])
        def add_one(feature):
            return feature + 1

        odt1 = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=add_one,
            transformation_type=TransformationType.ON_DEMAND,
        )

        odt2 = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=add_one,
            transformation_type=TransformationType.ON_DEMAND,
        )

        assert odt1 == odt2

    @pytest.mark.skipif(
        Version("4.1.6") >= Version(version.__version__),
        reason="Requires Hopsworks 4.1.7 or higher to be working.",
    )
    def test_inequality(self, mocker):
        mocked_connection = mocker.MagicMock()
        mocked_connection.backend_version = (
            version.__version__
        )  # Mocking backend version to be the same as the current version
        mocked_connection = mocker.patch(
            "hopsworks_common.client.get_connection", return_value=mocked_connection
        )

        @udf([int])
        def add_one(feature):
            return feature + 1

        mdt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=add_one,
            transformation_type=TransformationType.MODEL_DEPENDENT,
        )

        odt = TransformationFunction(
            featurestore_id=10,
            hopsworks_udf=add_one,
            transformation_type=TransformationType.ON_DEMAND,
        )

        assert mdt != odt



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json
from typing import Optional

import humps
from hopsworks_common import client, util
from hopsworks_common.core import (
    dataset_api,
    environment_api,
    flink_cluster_api,
    git_api,
    job_api,
    kafka_api,
    opensearch_api,
)


class Project:
    def __init__(
        self,
        archived=None,
        created=None,
        description=None,
        docker_image=None,
        hops_examples=None,
        inodeid=None,
        is_old_docker_image=None,
        is_preinstalled_docker_image=None,
        owner=None,
        project_id=None,
        project_name=None,
        project_team=None,
        quotas=None,
        retention_period=None,
        services=None,
        datasets=None,
        creation_status=None,
        project_namespace=None,
        **kwargs,
    ):
        self._id = project_id
        self._name = project_name
        self._owner = owner
        self._description = description
        self._created = created

        self._opensearch_api = opensearch_api.OpenSearchApi()
        self._kafka_api = kafka_api.KafkaApi()
        self._job_api = job_api.JobApi()
        self._jobs_api = self._job_api  # deprecated
        self._flink_cluster_api = flink_cluster_api.FlinkClusterApi()
        self._git_api = git_api.GitApi()
        self._dataset_api = dataset_api.DatasetApi()
        self._environment_api = environment_api.EnvironmentApi()
        self._project_namespace = project_namespace

    @classmethod
    def from_response_json(cls, json_dict):
        if json_dict:
            json_decamelized = humps.decamelize(json_dict)
            return cls(**json_decamelized)
        else:
            return None

    @property
    def id(self):
        """Id of the project"""
        return self._id

    @property
    def name(self):
        """Name of the project"""
        return self._name

    @property
    def owner(self):
        """Owner of the project"""
        return self._owner

    @property
    def description(self):
        """Description of the project"""
        return self._description

    @property
    def created(self):
        """Timestamp when the project was created"""
        return self._created

    @property
    def project_namespace(self):
        """Kubernetes namespace used by project"""
        return self._project_namespace

    def get_feature_store(
        self, name: Optional[str] = None
    ):  # -> hsfs.feature_store.FeatureStore
        """Connect to Project's Feature Store.

        Defaulting to the project name of default feature store. To get a
        shared feature store, the project name of the feature store is required.

        !!! example "Example for getting the Feature Store API of a project"
            ```python
            import hopsworks

            project = hopsworks.login()

            fs = project.get_feature_store()
            ```

        # Arguments
            name: Project name of the feature store.
        # Returns
            `hsfs.feature_store.FeatureStore`: The Feature Store API
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        return client.get_connection().get_feature_store(name)

    def get_model_registry(self):
        """Connect to Project's Model Registry API.

        !!! example "Example for getting the Model Registry API of a project"
            ```python
            import hopsworks

            project = hopsworks.login()

            mr = project.get_model_registry()
            ```

        # Returns
            `hsml.model_registry.ModelRegistry`: The Model Registry API
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        return client.get_connection().get_model_registry()

    def get_model_serving(self):
        """Connect to Project's Model Serving API.

        !!! example "Example for getting the Model Serving API of a project"
            ```python
            import hopsworks

            project = hopsworks.login()

            ms = project.get_model_serving()
            ```

        # Returns
            `hsml.model_serving.ModelServing`: The Model Serving API
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        return client.get_connection().get_model_serving()

    def get_kafka_api(self):
        """Get the kafka api for the project.

        # Returns
            `KafkaApi`: The Kafka Api handle
        """
        _client = client.get_instance()
        if _client._is_external():
            _client.download_certs()
        return self._kafka_api

    def get_opensearch_api(self):
        """Get the opensearch api for the project.

        # Returns
            `OpenSearchApi`: The OpenSearch Api handle
        """
        _client = client.get_instance()
        if _client._is_external():
            _client.download_certs()
        return self._opensearch_api

    def get_job_api(self):
        """Get the job API for the project.

        # Returns
            `JobApi`: The Job Api handle
        """
        return self._job_api

    def get_jobs_api(self):
        """**Deprecated**, use get_job_api instead. Excluded from docs to prevent API breakage"""
        return self.get_job_api()

    def get_flink_cluster_api(self):
        """Get the flink cluster API for the project.

        # Returns
            `FlinkClusterApi`: The Flink Cluster Api handle
        """
        return self._flink_cluster_api

    def get_git_api(self):
        """Get the git repository api for the project.

        # Returns
            `GitApi`: The Git Api handle
        """
        return self._git_api

    def get_dataset_api(self):
        """Get the dataset api for the project.

        # Returns
            `DatasetApi`: The Datasets Api handle
        """
        return self._dataset_api

    def get_environment_api(self):
        """Get the Python environment AP

        # Returns
            `EnvironmentApi`: The Python Environment Api handle
        """
        return self._environment_api

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def __str__(self):
        return self.json()

    def __repr__(self):
        if self._description is not None:
            return f"Project({self._name!r}, {self._owner!r}, {self._description!r})"
        else:
            return f"Project({self._name!r}, {self._owner!r})"

    def get_url(self):
        path = "/p/" + str(self.id)
        return util.get_hostname_replaced_url(path)



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import numpy as np
from hsml import schema


class TestSchema:
    # constructor

    def test_constructor_default(self, mocker):
        # Arrange
        mock_tensor = mocker.MagicMock()
        mock_tensor.tensors = mocker.MagicMock(return_value="tensor_schema")
        mock_columnar = mocker.MagicMock()
        mock_columnar.columns = mocker.MagicMock(return_value="columnar_schema")
        mock_convert_tensor_to_schema = mocker.patch(
            "hsml.schema.Schema._convert_tensor_to_schema", return_value=mock_tensor
        )
        mock_convert_columnar_to_schema = mocker.patch(
            "hsml.schema.Schema._convert_columnar_to_schema",
            return_value=mock_columnar,
        )

        # Act
        s = schema.Schema()

        # Assert
        assert s.columnar_schema == mock_columnar.columns
        assert not hasattr(s, "tensor_schema")
        mock_convert_tensor_to_schema.assert_not_called()
        mock_convert_columnar_to_schema.assert_called_once_with(None)

    def test_constructor_numpy(self, mocker):
        # Arrange
        obj = np.array([])
        mock_tensor = mocker.MagicMock()
        mock_tensor.tensors = mocker.MagicMock(return_value="tensor_schema")
        mock_columnar = mocker.MagicMock()
        mock_columnar.columns = mocker.MagicMock(return_value="columnar_schema")
        mock_convert_tensor_to_schema = mocker.patch(
            "hsml.schema.Schema._convert_tensor_to_schema", return_value=mock_tensor
        )
        mock_convert_columnar_to_schema = mocker.patch(
            "hsml.schema.Schema._convert_columnar_to_schema",
            return_value=mock_columnar,
        )

        # Act
        s = schema.Schema(obj)

        # Assert
        assert s.tensor_schema == mock_tensor.tensors
        assert not hasattr(s, "columnar_schema")
        mock_convert_columnar_to_schema.assert_not_called()
        mock_convert_tensor_to_schema.assert_called_once_with(obj)

    def test_constructor_tensor_list(self, mocker):
        # Arrange
        obj = [{"shape": "some_shape"}]
        mock_tensor = mocker.MagicMock()
        mock_tensor.tensors = mocker.MagicMock(return_value="tensor_schema")
        mock_columnar = mocker.MagicMock()
        mock_columnar.columns = mocker.MagicMock(return_value="columnar_schema")
        mock_convert_tensor_to_schema = mocker.patch(
            "hsml.schema.Schema._convert_tensor_to_schema", return_value=mock_tensor
        )
        mock_convert_columnar_to_schema = mocker.patch(
            "hsml.schema.Schema._convert_columnar_to_schema",
            return_value=mock_columnar,
        )

        # Act
        s = schema.Schema(obj)

        # Assert
        assert s.tensor_schema == mock_tensor.tensors
        assert not hasattr(s, "columnar_schema")
        mock_convert_columnar_to_schema.assert_not_called()
        mock_convert_tensor_to_schema.assert_called_once_with(obj)

    def test_constructor_column_list(self, mocker):
        # Arrange
        obj = [{"no_shape": "nothing"}]
        mock_tensor = mocker.MagicMock()
        mock_tensor.tensors = mocker.MagicMock(return_value="tensor_schema")
        mock_columnar = mocker.MagicMock()
        mock_columnar.columns = mocker.MagicMock(return_value="columnar_schema")
        mock_convert_tensor_to_schema = mocker.patch(
            "hsml.schema.Schema._convert_tensor_to_schema", return_value=mock_tensor
        )
        mock_convert_columnar_to_schema = mocker.patch(
            "hsml.schema.Schema._convert_columnar_to_schema",
            return_value=mock_columnar,
        )

        # Act
        s = schema.Schema(obj)

        # Assert
        assert s.columnar_schema == mock_columnar.columns
        assert not hasattr(s, "tensor_schema")
        mock_convert_tensor_to_schema.assert_not_called()
        mock_convert_columnar_to_schema.assert_called_once_with(obj)

    # convert to schema

    def test_convert_columnar_to_schema(self, mocker):
        # Arrange
        obj = {"key": "value"}
        mock_columnar_schema_init = mocker.patch(
            "hsml.utils.schema.columnar_schema.ColumnarSchema.__init__",
            return_value=None,
        )
        mock_schema = mocker.MagicMock()
        mock_schema._convert_columnar_to_schema = (
            schema.Schema._convert_columnar_to_schema
        )

        # Act
        ret = mock_schema._convert_columnar_to_schema(mock_schema, obj)

        # Assert
        assert isinstance(ret, schema.ColumnarSchema)
        mock_columnar_schema_init.assert_called_once_with(obj)

    def test_convert_tensor_to_schema(self, mocker):
        # Arrange
        obj = {"key": "value"}
        mock_tensor_schema_init = mocker.patch(
            "hsml.utils.schema.tensor_schema.TensorSchema.__init__",
            return_value=None,
        )
        mock_schema = mocker.MagicMock()
        mock_schema._convert_tensor_to_schema = schema.Schema._convert_tensor_to_schema

        # Act
        ret = mock_schema._convert_tensor_to_schema(mock_schema, obj)

        # Assert
        assert isinstance(ret, schema.TensorSchema)
        mock_tensor_schema_init.assert_called_once_with(obj)

    # get type

    def test_get_type_none(self, mocker):
        # Arrange
        class MockSchema:
            pass

        mock_schema = MockSchema()
        mock_schema._get_type = schema.Schema._get_type

        # Act
        t = mock_schema._get_type(mock_schema)

        # Assert
        assert t is None

    def test_get_type_tensor(self, mocker):
        # Arrange
        class MockSchema:
            tensor_schema = {}

        mock_schema = MockSchema()
        mock_schema._get_type = schema.Schema._get_type

        # Act
        t = mock_schema._get_type(mock_schema)

        # Assert
        assert t == "tensor"

    def test_get_type_columnar(self, mocker):
        # Arrange
        class MockSchema:
            columnar_schema = {}

        mock_schema = MockSchema()
        mock_schema._get_type = schema.Schema._get_type

        # Act
        t = mock_schema._get_type(mock_schema)

        # Assert
        assert t == "columnar"



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import copy

import pytest
from hsml import resources, transformer
from hsml.constants import RESOURCES


SERVING_RESOURCE_LIMITS = {"cores": 2, "memory": 1024, "gpus": 2}
SERVING_NUM_INSTANCES_NO_LIMIT = [-1]
SERVING_NUM_INSTANCES_SCALE_TO_ZERO = [0]
SERVING_NUM_INSTANCES_ONE = [0]


class TestTransformer:
    # from response json

    def test_from_response_json_with_transformer_field(self, mocker, backend_fixtures):
        # Arrange
        self._mock_serving_variables(mocker, SERVING_NUM_INSTANCES_NO_LIMIT)
        json = backend_fixtures["transformer"]["get_deployment_with_transformer"][
            "response"
        ]

        # Act
        t = transformer.Transformer.from_response_json(json)

        # Assert
        assert isinstance(t, transformer.Transformer)
        assert t.script_file == json["transformer"]

        tr_resources = json["transformer_resources"]
        assert (
            t.resources.num_instances == tr_resources["requested_transformer_instances"]
        )
        assert t.resources.requests.cores == tr_resources["requests"]["cores"]
        assert t.resources.requests.memory == tr_resources["requests"]["memory"]
        assert t.resources.requests.gpus == tr_resources["requests"]["gpus"]
        assert t.resources.limits.cores == tr_resources["limits"]["cores"]
        assert t.resources.limits.memory == tr_resources["limits"]["memory"]
        assert t.resources.limits.gpus == tr_resources["limits"]["gpus"]

    def test_from_response_json_with_script_file_field(self, mocker, backend_fixtures):
        # Arrange
        self._mock_serving_variables(mocker, SERVING_NUM_INSTANCES_NO_LIMIT)
        json = backend_fixtures["transformer"]["get_transformer_with_resources"][
            "response"
        ]

        # Act
        t = transformer.Transformer.from_response_json(json)

        # Assert
        assert isinstance(t, transformer.Transformer)
        assert t.script_file == json["script_file"]

        tr_resources = json["resources"]
        assert t.resources.num_instances == tr_resources["num_instances"]
        assert t.resources.requests.cores == tr_resources["requests"]["cores"]
        assert t.resources.requests.memory == tr_resources["requests"]["memory"]
        assert t.resources.requests.gpus == tr_resources["requests"]["gpus"]
        assert t.resources.limits.cores == tr_resources["limits"]["cores"]
        assert t.resources.limits.memory == tr_resources["limits"]["memory"]
        assert t.resources.limits.gpus == tr_resources["limits"]["gpus"]

    def test_from_response_json_empty(self, mocker, backend_fixtures):
        # Arrange
        self._mock_serving_variables(mocker, SERVING_NUM_INSTANCES_NO_LIMIT)
        json = backend_fixtures["transformer"]["get_deployment_without_transformer"][
            "response"
        ]

        # Act
        t = transformer.Transformer.from_response_json(json)

        # Assert
        assert t is None

    def test_from_response_json_default_resources(self, mocker, backend_fixtures):
        # Arrange
        self._mock_serving_variables(
            mocker, SERVING_NUM_INSTANCES_NO_LIMIT, force_scale_to_zero=False
        )
        json = backend_fixtures["transformer"]["get_transformer_without_resources"][
            "response"
        ]

        # Act
        t = transformer.Transformer.from_response_json(json)

        # Assert
        assert isinstance(t, transformer.Transformer)
        assert t.script_file == json["script_file"]

        assert t.resources.num_instances == RESOURCES.MIN_NUM_INSTANCES
        assert t.resources.requests.cores == RESOURCES.MIN_CORES
        assert t.resources.requests.memory == RESOURCES.MIN_MEMORY
        assert t.resources.requests.gpus == RESOURCES.MIN_GPUS
        assert t.resources.limits.cores == RESOURCES.MAX_CORES
        assert t.resources.limits.memory == RESOURCES.MAX_MEMORY
        assert t.resources.limits.gpus == RESOURCES.MAX_GPUS

    # constructor

    def test_constructor_default_resources(self, mocker, backend_fixtures):
        # Arrange
        self._mock_serving_variables(
            mocker, SERVING_NUM_INSTANCES_NO_LIMIT, force_scale_to_zero=False
        )
        json = backend_fixtures["transformer"]["get_transformer_without_resources"][
            "response"
        ]

        # Act
        t = transformer.Transformer(json["script_file"], resources=None)

        # Assert
        assert t.script_file == json["script_file"]

        assert t.resources.num_instances == RESOURCES.MIN_NUM_INSTANCES
        assert t.resources.requests.cores == RESOURCES.MIN_CORES
        assert t.resources.requests.memory == RESOURCES.MIN_MEMORY
        assert t.resources.requests.gpus == RESOURCES.MIN_GPUS
        assert t.resources.limits.cores == RESOURCES.MAX_CORES
        assert t.resources.limits.memory == RESOURCES.MAX_MEMORY
        assert t.resources.limits.gpus == RESOURCES.MAX_GPUS

    def test_constructor_default_resources_when_scale_to_zero_is_required(
        self, mocker, backend_fixtures
    ):
        # Arrange
        self._mock_serving_variables(
            mocker, SERVING_NUM_INSTANCES_NO_LIMIT, force_scale_to_zero=True
        )
        json = backend_fixtures["transformer"]["get_transformer_without_resources"][
            "response"
        ]

        # Act
        t = transformer.Transformer(json["script_file"], resources=None)

        # Assert
        assert t.script_file == json["script_file"]

        assert t.resources.num_instances == 0
        assert t.resources.requests.cores == RESOURCES.MIN_CORES
        assert t.resources.requests.memory == RESOURCES.MIN_MEMORY
        assert t.resources.requests.gpus == RESOURCES.MIN_GPUS
        assert t.resources.limits.cores == RESOURCES.MAX_CORES
        assert t.resources.limits.memory == RESOURCES.MAX_MEMORY
        assert t.resources.limits.gpus == RESOURCES.MAX_GPUS

    # validate resources

    def test_validate_resources_none(self):
        # Act
        res = transformer.Transformer._validate_resources(None)

        # Assert
        assert res is None

    def test_validate_resources_num_instances_zero(self, mocker):
        # Arrange
        self._mock_serving_variables(
            mocker, SERVING_NUM_INSTANCES_NO_LIMIT, force_scale_to_zero=False
        )
        tr = resources.TransformerResources(num_instances=0)

        # Act
        res = transformer.Transformer._validate_resources(tr)

        # Assert
        assert res == tr

    def test_validate_resources_num_instances_one_without_scale_to_zero(self, mocker):
        # Arrange
        self._mock_serving_variables(
            mocker, SERVING_NUM_INSTANCES_NO_LIMIT, force_scale_to_zero=False
        )
        tr = resources.TransformerResources(num_instances=1)

        # Act
        res = transformer.Transformer._validate_resources(tr)

        # Assert
        assert res == tr

    def test_validate_resources_num_instances_one_with_scale_to_zero(self, mocker):
        # Arrange
        self._mock_serving_variables(
            mocker, SERVING_NUM_INSTANCES_NO_LIMIT, force_scale_to_zero=True
        )
        tr = resources.TransformerResources(num_instances=1)

        # Act
        with pytest.raises(ValueError) as e_info:
            _ = transformer.Transformer._validate_resources(tr)

        # Assert
        assert "Scale-to-zero is required" in str(e_info.value)

    # default num instances

    def test_get_default_num_instances_without_scale_to_zero(self, mocker):
        # Arrange
        self._mock_serving_variables(
            mocker, SERVING_NUM_INSTANCES_NO_LIMIT, force_scale_to_zero=False
        )

        # Act
        num_instances = transformer.Transformer._get_default_num_instances()

        # Assert
        assert num_instances == RESOURCES.MIN_NUM_INSTANCES

    def test_get_default_num_instances_with_scale_to_zero(self, mocker):
        # Arrange
        self._mock_serving_variables(
            mocker, SERVING_NUM_INSTANCES_NO_LIMIT, force_scale_to_zero=True
        )

        # Act
        num_instances = transformer.Transformer._get_default_num_instances()

        # Assert
        assert num_instances == 0

    # default resources

    def test_get_default_resources_without_scale_to_zero(self, mocker):
        # Arrange
        self._mock_serving_variables(
            mocker, SERVING_NUM_INSTANCES_NO_LIMIT, force_scale_to_zero=False
        )

        # Act
        res = transformer.Transformer._get_default_resources()

        # Assert
        assert isinstance(res, resources.TransformerResources)
        assert res.num_instances == RESOURCES.MIN_NUM_INSTANCES

    def test_get_default_resources_with_scale_to_zero(self, mocker):
        # Arrange
        self._mock_serving_variables(
            mocker, SERVING_NUM_INSTANCES_NO_LIMIT, force_scale_to_zero=True
        )

        # Act
        res = transformer.Transformer._get_default_resources()

        # Assert
        assert isinstance(res, resources.TransformerResources)
        assert res.num_instances == 0

    # extract fields from json

    def test_extract_fields_from_json(self, mocker, backend_fixtures):
        # Arrange
        self._mock_serving_variables(mocker, SERVING_NUM_INSTANCES_NO_LIMIT)
        json = backend_fixtures["transformer"]["get_deployment_with_transformer"][
            "response"
        ]
        json_copy = copy.deepcopy(json)

        # Act
        sf, rc = transformer.Transformer.extract_fields_from_json(json_copy)

        # Assert
        assert sf == json["transformer"]
        assert isinstance(rc, resources.TransformerResources)

        tr_resources = json["transformer_resources"]
        assert rc.num_instances == tr_resources["requested_transformer_instances"]
        assert rc.requests.cores == tr_resources["requests"]["cores"]
        assert rc.requests.memory == tr_resources["requests"]["memory"]
        assert rc.requests.gpus == tr_resources["requests"]["gpus"]
        assert rc.limits.cores == tr_resources["limits"]["cores"]
        assert rc.limits.memory == tr_resources["limits"]["memory"]
        assert rc.limits.gpus == tr_resources["limits"]["gpus"]

    # auxiliary methods

    def _mock_serving_variables(self, mocker, num_instances, force_scale_to_zero=False):
        mocker.patch(
            "hopsworks_common.client.get_serving_resource_limits",
            return_value=SERVING_RESOURCE_LIMITS,
        )
        mocker.patch(
            "hopsworks_common.client.get_serving_num_instances_limits",
            return_value=num_instances,
        )
        mocker.patch(
            "hopsworks_common.client.is_scale_to_zero_required",
            return_value=force_scale_to_zero,
        )



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from hopsworks_common.usage import (
    HOPSWORKS_DIR,
    EnvironmentAttribute,
    MethodCounter,
    disable,
    enable,
    get_env,
    init_usage,
    method_logger,
)


__all__ = [
    "HOPSWORKS_DIR",
    "EnvironmentAttribute",
    "MethodCounter",
    "disable",
    "enable",
    "get_env",
    "init_usage",
    "method_logger",
]



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from typing import Dict, Optional

from hopsworks_common import client
from hsfs import feature_group as fg_mod
from hsfs.core import online_ingestion


class OnlineIngestionApi:
    def create_online_ingestion(
        self,
        feature_group_instance: fg_mod.FeatureGroup,
        online_ingestion_instance: online_ingestion.OnlineIngestion,
    ) -> online_ingestion.OnlineIngestion:
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_group_instance.feature_store_id,
            "featuregroups",
            feature_group_instance.id,
            "online_ingestion",
        ]

        headers = {"content-type": "application/json"}
        return online_ingestion.OnlineIngestion.from_response_json(
            _client._send_request(
                "POST",
                path_params,
                headers=headers,
                data=online_ingestion_instance.json(),
            ),
            feature_group=feature_group_instance,
        )

    def get_online_ingestion(
        self,
        feature_group_instance: fg_mod.FeatureGroup,
        query_params: Optional[Dict[str, str]] = None,
    ) -> online_ingestion.OnlineIngestion:
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_group_instance.feature_store_id,
            "featuregroups",
            feature_group_instance.id,
            "online_ingestion",
        ]

        return online_ingestion.OnlineIngestion.from_response_json(
            _client._send_request("GET", path_params, query_params),
            feature_group=feature_group_instance,
        )



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import getpass
import logging
import os
import sys
import tempfile
import warnings
from pathlib import Path
from typing import Literal, Optional, Union

from hopsworks import client, constants, project, version
from hopsworks.client.exceptions import (
    HopsworksSSLClientError,
    ProjectException,
    RestAPIError,
)
from hopsworks.connection import Connection
from hopsworks.core import project_api, secret_api
from hopsworks.decorators import NoHopsworksConnectionError
from hopsworks_common import usage
from requests.exceptions import SSLError


# Needs to run before import of hsml and hsfs
warnings.filterwarnings(action="ignore", category=UserWarning, module=r".*psycopg2")

import hsfs  # noqa: E402
import hsml  # noqa: E402


sys.modules["hopsworks.hsfs"] = hsfs
sys.modules["hopsworks.hsml"] = hsml


__version__ = version.__version__

connection = Connection.connection

_hw_connection = Connection.connection

_connected_project = None
_secrets_api = None
_project_api = None

udf = hsfs.hopsworks_udf.udf


def hw_formatwarning(message, category, filename, lineno, line=None):
    return "{}: {}\n".format(category.__name__, message)


warnings.formatwarning = hw_formatwarning

__all__ = ["connection", "udf"]

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)s: %(message)s",
    stream=sys.stdout,
)


def login(
    host: Optional[str] = None,
    port: int = 443,
    project: Optional[str] = None,
    api_key_value: Optional[str] = None,
    api_key_file: Optional[str] = None,
    hostname_verification: bool = False,
    trust_store_path: Optional[str] = None,
    engine: Union[
        None,
        Literal["spark"],
        Literal["python"],
        Literal["training"],
        Literal["spark-no-metastore"],
        Literal["spark-delta"],
    ] = None,
) -> project.Project:
    """Connect to [Serverless Hopsworks](https://app.hopsworks.ai) by calling the `hopsworks.login()` function with no arguments.

    !!! example "Connect to Serverless"
        ```python

        import hopsworks

        project = hopsworks.login()

        ```

    Alternatively, connect to your own Hopsworks installation by specifying the host, port and api key.

    !!! example "Connect to your Hopsworks cluster"
        ```python

        import hopsworks

        project = hopsworks.login(host="my.hopsworks.server",
                                  port=8181,
                                  api_key_value="DKN8DndwaAjdf98FFNSxwdVKx")

        ```

    In addition to setting function arguments directly, `hopsworks.login()` also reads the environment variables:
    HOPSWORKS_HOST, HOPSWORKS_PORT, HOPSWORKS_PROJECT, HOPSWORKS_API_KEY, HOPSWORKS_HOSTNAME_VERIFICATION, HOPSWORKS_TRUST_STORE_PATH and HOPSWORKS_ENGINE.

    The function arguments do however take precedence over the environment variables in case both are set.

    # Arguments
        host: The hostname of the Hopsworks instance, defaults to `None`.
        port: The port on which the Hopsworks instance can be reached,
            defaults to `443`.
        project: Name of the project to access. If used inside a Hopsworks environment it always gets the current project. If not provided you will be prompted to enter it.
        api_key_value: Value of the Api Key
        api_key_file: Path to file wih Api Key
        hostname_verification: Whether to verify Hopsworks' certificate
        trust_store_path: Path on the file system containing the Hopsworks certificates
        engine: Specifies the engine to use. Possible options are "spark", "python", "training", "spark-no-metastore", or "spark-delta". The default value is None, which automatically selects the engine based on the environment:
            "spark": Used if Spark is available, such as in Hopsworks or Databricks environments.
            "python": Used in local Python environments or AWS SageMaker when Spark is not available.
            "training": Used when only feature store metadata is needed, such as for obtaining training dataset locations and label information during Hopsworks training experiments.
            "spark-no-metastore": Functions like "spark" but does not rely on the Hive metastore.
            "spark-delta": Minimizes dependencies further by avoiding both Hive metastore and HopsFS.
    # Returns
        `Project`: The Project object to perform operations on
    # Raises
        `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        `hopsworks.client.exceptions.HopsworksSSLClientError`: If SSLError is raised from underlying requests library
    """

    global _connected_project

    # If already logged in, should reset connection and follow login procedure as Connection may no longer be valid
    logout()

    global _hw_connection

    # If project argument not defined, get HOPSWORKS_ENGINE environment variable
    if engine is None and "HOPSWORKS_ENGINE" in os.environ:
        engine = os.environ["HOPSWORKS_ENGINE"]

    # If inside hopsworks, just return the current project for now
    if "REST_ENDPOINT" in os.environ:
        _hw_connection = _hw_connection(
            hostname_verification=hostname_verification, engine=engine
        )
        _connected_project = _hw_connection.get_project()
        _initialize_module_apis()
        print("\nLogged in to project, explore it here " + _connected_project.get_url())
        return _connected_project

    # This is run for an external client

    # Function arguments takes precedence over environment variable.
    # Here we check if environment variable exists and function argument is not set, we use the environment variable.

    # If api_key_value/api_key_file not defined, get HOPSWORKS_API_KEY environment variable
    api_key = None
    if (
        api_key_value is None
        and api_key_file is None
        and "HOPSWORKS_API_KEY" in os.environ
    ):
        api_key = os.environ["HOPSWORKS_API_KEY"]

    # If project argument not defined, get HOPSWORKS_PROJECT environment variable
    if project is None and "HOPSWORKS_PROJECT" in os.environ:
        project = os.environ["HOPSWORKS_PROJECT"]

    # If host argument not defined, get HOPSWORKS_HOST environment variable
    if host is None and "HOPSWORKS_HOST" in os.environ:
        host = os.environ["HOPSWORKS_HOST"]
    elif host is None:  # Always do a fallback to Serverless Hopsworks if not defined
        host = constants.HOSTS.APP_HOST

    is_app = host == constants.HOSTS.APP_HOST

    # If port same as default, get HOPSWORKS_HOST environment variable
    if port == 443 and "HOPSWORKS_PORT" in os.environ:
        port = os.environ["HOPSWORKS_PORT"]

    hostname_verification = os.getenv(
        "HOPSWORKS_HOSTNAME_VERIFICATION", "{}".format(hostname_verification)
    ).lower() in ("true", "1", "y", "yes")

    trust_store_path = os.getenv("HOPSWORKS_TRUST_STORE_PATH", trust_store_path)

    # This .hw_api_key is created when a user logs into Serverless Hopsworks the first time.
    # It is then used only for future login calls to Serverless. For other Hopsworks installations it's ignored.
    api_key_path = _get_cached_api_key_path()

    # Conditions for getting the api_key
    # If user supplied the api key directly
    if api_key_value is not None:
        api_key = api_key_value
    # If user supplied the api key in a file
    elif api_key_file is not None:
        file = None
        if os.path.exists(api_key_file):
            try:
                file = open(api_key_file, mode="r")
                api_key = file.read()
            finally:
                file.close()
        else:
            raise IOError(
                "Could not find api key file on path: {}".format(api_key_file)
            )
    # If user connected to Serverless Hopsworks, and the cached .hw_api_key exists, then use it.
    elif os.path.exists(api_key_path) and is_app:
        try:
            _hw_connection = _hw_connection(
                host=host,
                port=port,
                engine=engine,
                api_key_file=api_key_path,
                hostname_verification=hostname_verification,
                trust_store_path=trust_store_path,
            )
            _connected_project = _prompt_project(_hw_connection, project, is_app)
            if _connected_project:
                _set_active_project(_connected_project)
            print(
                "\nLogged in to project, explore it here "
                + _connected_project.get_url()
            )
            _initialize_module_apis()
            return _connected_project
        except RestAPIError:
            logout()
            # API Key may be invalid, have the user supply it again
            os.remove(api_key_path)
        except SSLError as ssl_e:
            logout()
            _handle_ssl_errors(ssl_e)

    if api_key is None and is_app:
        print(
            "Copy your Api Key (first register/login): https://c.app.hopsworks.ai/account/api/generated"
        )
        api_key = getpass.getpass(prompt="\nPaste it here: ")

        # If api key was provided as input, save the API key locally on disk to avoid users having to enter it again in the same environment
        Path(os.path.dirname(api_key_path)).mkdir(parents=True, exist_ok=True)
        descriptor = os.open(
            path=api_key_path, flags=(os.O_WRONLY | os.O_CREAT | os.O_TRUNC), mode=0o600
        )
        with open(descriptor, "w") as fh:
            fh.write(api_key.strip())

    try:
        _hw_connection = _hw_connection(
            host=host,
            port=port,
            engine=engine,
            api_key_value=api_key,
            hostname_verification=hostname_verification,
            trust_store_path=trust_store_path,
        )
        _connected_project = _prompt_project(_hw_connection, project, is_app)
        if _connected_project:
            _set_active_project(_connected_project)
    except RestAPIError as hw_e:
        logout()
        raise hw_e
    except SSLError as ssl_e:
        logout()
        _handle_ssl_errors(ssl_e)

    if _connected_project is None:
        print(
            "Could not find any project, use hopsworks.create_project('my_project') to create one"
        )
    else:
        print("\nLogged in to project, explore it here " + _connected_project.get_url())

    _initialize_module_apis()
    return _connected_project


def _handle_ssl_errors(ssl_e):
    raise HopsworksSSLClientError(
        "Hopsworks certificate verification can be turned off by specifying hopsworks.login(hostname_verification=False) "
        "or setting the environment variable HOPSWORKS_HOSTNAME_VERIFICATION='False'"
    ) from ssl_e


def _get_cached_api_key_path():
    """
    This function is used to get an appropriate path to store the user supplied API Key for Serverless Hopsworks.

    First it will search for .hw_api_key in the current working directory, if it exists it will use it (this is default in 3.0 client)
    Otherwise, falls back to storing the API key in HOME
    If not sufficient permissions are set in HOME to create the API key (writable and executable), it uses the temp directory to store it.

    """

    api_key_name = ".hw_api_key"
    api_key_folder = ".{}_hopsworks_app".format(getpass.getuser())

    # Path for current working directory api key
    cwd_api_key_path = os.path.join(os.getcwd(), api_key_name)

    # Path for home api key
    home_dir_path = os.path.expanduser("~")
    home_api_key_path = os.path.join(home_dir_path, api_key_folder, api_key_name)

    # Path for tmp api key
    temp_dir_path = tempfile.gettempdir()
    temp_api_key_path = os.path.join(temp_dir_path, api_key_folder, api_key_name)

    if os.path.exists(
        cwd_api_key_path
    ):  # For backward compatibility, if .hw_api_key exists in current working directory get it from there
        api_key_path = cwd_api_key_path
    elif os.access(home_dir_path, os.W_OK) and os.access(
        home_dir_path, os.X_OK
    ):  # Otherwise use the home directory of the user
        api_key_path = home_api_key_path
    else:  # Finally try to store it in temp
        api_key_path = temp_api_key_path

    return api_key_path


def _prompt_project(valid_connection, project, is_app):
    if project is None:
        if is_app:
            # On Serverless we filter out projects owned by other users to make sure automatic login
            # without a prompt still happens when users add showcase projects created by other users
            saas_projects = valid_connection._project_api._get_owned_projects()
        else:
            saas_projects = valid_connection._project_api._get_projects()
        if len(saas_projects) == 0:
            if is_app:
                raise ProjectException("Could not find any project")
            else:
                return None
        elif len(saas_projects) == 1:
            return saas_projects[0]
        else:
            while True:
                print("\nMultiple projects found. \n")
                for index in range(len(saas_projects)):
                    print("\t (" + str(index + 1) + ") " + saas_projects[index].name)
                while True:
                    project_index = input(
                        "\nEnter number corresponding to the project to use: "
                    )
                    # Handle invalid input type
                    try:
                        project_index = int(project_index)
                        # Handle negative indexing
                        if project_index <= 0:
                            print("Invalid input, must be greater than or equal to 1")
                            continue
                        # Handle index out of range
                        try:
                            return saas_projects[project_index - 1]
                        except IndexError:
                            print(
                                "Invalid input, should be an integer from the list of projects."
                            )
                    except ValueError:
                        print(
                            "Invalid input, should be an integer from the list of projects."
                        )
    else:
        try:
            return valid_connection._project_api._get_project(project)
        except RestAPIError as x:
            raise ProjectException("Could not find project {}".format(project)) from x


def logout():
    """Cleans up and closes the connection for hopsworks."""
    global _hw_connection
    global _project_api
    global _secrets_api

    if _is_connection_active():
        _hw_connection.close()

    client.stop()
    _project_api = None
    _secrets_api = None
    _hw_connection = Connection.connection


def _is_connection_active():
    global _hw_connection
    return isinstance(_hw_connection, Connection)


def get_current_project() -> project.Project:
    """Get a reference to the current logged in project.

    !!! example "Example for getting the project reference"
        ```python

        import hopsworks

        hopsworks.login()

        project = hopsworks.get_current_project()

        ```

    # Returns
        `Project`. The Project object to perform operations on
    """
    global _connected_project
    if _connected_project is None:
        raise ProjectException("No project is set for this session")
    return _connected_project


def _initialize_module_apis():
    global _project_api
    global _secrets_api
    _project_api = project_api.ProjectApi()
    _secrets_api = secret_api.SecretsApi()


def create_project(name: str, description: Optional[str] = None, feature_store_topic: Optional[str] = None):
    """Create a new project.

    !!! warning "Not supported"
        This is not supported if you are connected to [Serverless Hopsworks](https://app.hopsworks.ai)

    !!! example "Example for creating a new project"
        ```python

        import hopsworks

        hopsworks.login(...)

        hopsworks.create_project("my_project", description="An example Hopsworks project")

        ```
    # Arguments
        name: The name of the project.
        description: optional description of the project
        feature_store_topic: optional feature store topic name

    # Returns
        `Project`. The Project object to perform operations on
    """
    global _hw_connection
    global _connected_project

    if not _is_connection_active():
        raise NoHopsworksConnectionError()

    new_project = _hw_connection._project_api._create_project(
        name, description, feature_store_topic
    )
    if _connected_project is None:
        _connected_project = new_project
        if _connected_project:
            _set_active_project(_connected_project)
        print(
            "Setting {} as the current project, a reference can be retrieved by calling hopsworks.get_current_project()".format(
                _connected_project.name
            )
        )
        return _connected_project
    else:
        print(
            "You are already using the project {}, to access the new project use hopsworks.login(..., project='{}')".format(
                _connected_project.name, new_project.name
            )
        )


def get_secrets_api():
    """Get the secrets api.

    # Returns
        `SecretsApi`: The Secrets Api handle
    """
    global _secrets_api
    if not _is_connection_active():
        raise NoHopsworksConnectionError()
    return _secrets_api


def _set_active_project(project):
    _client = client.get_instance()
    if _client._is_external():
        _client.provide_project(project.name)


def disable_usage_logging():
    usage.disable()


def get_sdk_info():
    return usage.get_env()
